// $Id: x_handler.cc,v 1.14 2003/02/09 19:28:00 gordon Exp $	
//
// Copyright (C) 2002 Gordon Fraser <gordon@debian.org>
// Parts copyright (C) 1998  Sam Hawker <shawkie@geocities.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// See the COPYING file for more details
//

#include "x_handler.h"

//--------------------------------------------------------------------
XHandler::XHandler(const string& display, const string& position, char x_flags)
  : window_size_(0), screen_(0), is_windowed_(x_flags & WINDOWED), 
    is_wmaker_(x_flags & WMAKER), is_ushape_(x_flags & USHAPED), 
    is_astep_(x_flags & ASTEP), display_(NULL), offset_w_(0), offset_h_(0),
    width_(PANEL_SIZE), height_(PANEL_SIZE), depth_(0)
{
  display_name_  = (char*)display.c_str();
  position_name_ = (char*)position.c_str();
  
  if(is_windowed_)
    is_wmaker_ = false;
  else
    is_wmaker_ = true;
}

//--------------------------------------------------------------------
XHandler::~XHandler()
{
  XFreeGC(display_, graphics_context_);
  XFreePixmap(display_, pixmap_display_);
  XDestroyWindow(display_, window_main_);

  if(is_wmaker_)
    XDestroyWindow(display_, window_icon_);

  XCloseDisplay(display_);

  delete pixmap_mask_;
  delete[] display_name_;
  delete[] position_name_;
}

//--------------------------------------------------------------------
void XHandler::setDisplay(const string& display_name)
{
  delete[] display_name_;
  display_name_ =(char*)display_name.c_str();
}

//--------------------------------------------------------------------
Display* XHandler::getDisplay() const
{
  return display_;
}

//--------------------------------------------------------------------
Window XHandler::getWindow() const
{
  return window_icon_;
}

//--------------------------------------------------------------------
void XHandler::setPosition(char* position)
{
  delete[] position_name_;

  position_name_ = new char[strlen(position)+1];
  strcpy(position_name_, position);  
}

//--------------------------------------------------------------------
void XHandler::setWindowMaker()
{
  is_wmaker_ = true;
}

//--------------------------------------------------------------------
void XHandler::setAfterStep()
{
  is_astep_ = true;
}

//--------------------------------------------------------------------
void XHandler::setUnshaped()
{
  is_ushape_ = true;
}


//--------------------------------------------------------------------
// TODO: check if everything is set
void XHandler::init(int argc, char** argv)
{
  window_size_ = is_astep_ ? ASTEPSIZE : NORMSIZE;
  
  if((display_ = XOpenDisplay(display_name_)) == NULL)
  {
    if(display_name_ == NULL)
      std::cerr <<  NAME << " : Unable to open X display " << std::endl;
    else
      std::cerr <<  NAME << " : Unable to open X display '" << XDisplayName(display_name_) << "'." << std::endl;
    exit(1);
  }
  
  initWindow(argc, argv);

  depth_ = DefaultDepth(display_, DefaultScreen(display_));
  //  pixmap_display_ = XCreatePixmap(display_, window_root_, PANEL_SIZE, PANEL_SIZE, display_depth);
  pixmap_display_ = XCreatePixmap(display_, window_root_, 
				  width_, height_, depth_);

  initGraphicsContext();

  XMapWindow(display_, window_main_);
}


//--------------------------------------------------------------------
void XHandler::initMask(Pixmap* mask, unsigned width, unsigned height)
{
  width_ = width;
  height_ = height;
  offset_w_ = (PANEL_SIZE - width) >> 1;
  offset_h_ = (PANEL_SIZE - height) >> 1;

  pixmap_mask_ = mask;

//   XFillRectangle(display_, pixmap_display_, graphics_context_, 
// 		 0, 0, width_, height_);

//    XResizeWindow(display_, window_icon_, width, height);
//    XResizeWindow(display_, window_main_, width, height);

//   Pixmap bg;
//   XCopyArea(display_, *mask, bg, graphics_context_, 0, 0, width, height,
//                   offset_w_, offset_w_);


  //  XSetWindowBackgroundPixmap(display_, window_icon_, bg);
//  XSetClipMask(display_, graphics_context_, *pixmap_mask_);
//  XSetClipMask(display_, graphics_context_, None);

  //    if(is_wmaker_ || is_ushape_ || is_astep_) 
  //      {
       
//    XShapeCombineMask(display_, window_icon_, ShapeBounding,
//    		    0, 0, *pixmap_mask_, ShapeSet);
   XShapeCombineMask(display_, window_main_, ShapeBounding,
  		    offset_w_, offset_h_, *pixmap_mask_, ShapeSet);
   XShapeCombineMask(display_, window_icon_, ShapeBounding,
     		    offset_w_, offset_h_, *pixmap_mask_, ShapeSet);
  
  //    XShapeOffsetShape(display_, window_icon_, ShapeBounding, 3,3);
//   XShapeCombineMask(display_, window_main_, ShapeBounding,
//     		    offset_w_, offset_h_, *pixmap_mask_, ShapeSet);
  //   //      } 
//   //    else 
//   //      {
//   //        XCopyArea(display_, pixmap_tile_, pixmap_display_, graphics_context_, 0, 0, 64, 64, 0, 0);
//   //      }

  //   XShapeCombineMask(display_, window_icon_, ShapeBounding,
  // 		    window_size_/2-32, window_size_/2-32, *pixmap_mask_, ShapeSet);
  //   XShapeCombineMask(display_, window_main_, ShapeBounding,
  // 		    window_size_/2-32, window_size_/2-32, *pixmap_mask_, ShapeSet);

  XSizeHints size_hints;
  XWMHints wmhints;

  size_hints.min_width   = size_hints.max_width  = width_;
  size_hints.min_height  = size_hints.max_height = height_;
  size_hints.width       = width_;
  size_hints.height      = height_;
  size_hints.x = offset_w_;
  size_hints.y = offset_h_;
  size_hints.flags = USSize | PMinSize | PMaxSize | PBaseSize;

  wmhints.icon_x = size_hints.x;
  wmhints.icon_y = size_hints.y;

  XSetWMNormalHints(display_, window_main_, &size_hints);
  if(is_wmaker_)
  {
    wmhints.initial_state = WithdrawnState;
    wmhints.flags = StateHint | IconWindowHint | IconPositionHint | WindowGroupHint;
    wmhints.window_group = window_main_;
    wmhints.icon_window = window_icon_;
      
    wmhints.icon_x = size_hints.x;
    wmhints.icon_y = size_hints.y;
  } 
  else 
  {
    wmhints.initial_state = NormalState;
    wmhints.flags = WindowGroupHint | StateHint;
  }

  XSetWMHints(display_, window_main_, &wmhints);  
}

//--------------------------------------------------------------------
void XHandler::initWindow(int argc, char** argv)
{
  char *wname = argv[0];
  int dummy = 0;
  XWMHints wmhints;
  XSizeHints size_hints;
  XClassHint class_hints;
  XTextProperty	name;

  screen_      = DefaultScreen(display_);
  window_root_ = RootWindow(display_, screen_);

  _XA_GNUSTEP_WM_FUNC = XInternAtom(display_, "_GNUSTEP_WM_FUNCTION", false);
  deleteWin = XInternAtom(display_, "WM_DELETE_WINDOW", false);


  size_hints.x = 0;
  size_hints.y = 0;
  size_hints.flags  = USSize;
  size_hints.flags  = 0; // Gordon
  
  bool pos = (XWMGeometry(display_, DefaultScreen(display_),
			  position_name_, NULL, 0, &size_hints, &size_hints.x, &size_hints.y,
			  &size_hints.width, &size_hints.height, &dummy)
	      & (XValue | YValue));
  
  size_hints.min_width   = size_hints.max_width  = width_;
  size_hints.min_height  = size_hints.max_height = height_;
  size_hints.width       = width_;
  size_hints.height      = height_;
  

  //   size_hints.min_width   = window_size_;
  //   size_hints.min_height  = window_size_;
  //   size_hints.max_width   = window_size_;
  //   size_hints.max_height  = window_size_;
  //   size_hints.base_width  = window_size_;
  //   size_hints.base_height = window_size_;
  //   size_hints.width       = window_size_;
  //   size_hints.height      = window_size_;
  size_hints.flags = PMinSize | PMaxSize | PBaseSize;


  back_pix_ = getColor("white");
  fore_pix_ = getColor("black");

  //   window_main_ = XCreateSimpleWindow(display_, window_root_, size_hints.x, size_hints.y,
  // 			    size_hints.width, size_hints.height, 0, fore_pix_, back_pix_);
  
  //   window_icon_ = XCreateSimpleWindow(display_, window_main_, size_hints.x, size_hints.y,
  // 				size_hints.width, size_hints.height, 0, fore_pix_, back_pix_);
  window_main_ = XCreateSimpleWindow(display_, window_root_, 0, 0,
				     width_, height_, 0, fore_pix_, back_pix_);
  
  window_icon_ = XCreateSimpleWindow(display_, window_main_, 0, 0,
				     width_, height_, 0, fore_pix_, back_pix_);
  XSetWMNormalHints(display_, window_main_, &size_hints);
  
  wmhints.icon_x = size_hints.x;
  wmhints.icon_y = size_hints.y;
  
  if(is_wmaker_ || is_astep_ || pos)
    size_hints.flags |= USPosition;

  if(is_wmaker_)
  {
    wmhints.initial_state = WithdrawnState;
    wmhints.flags = StateHint | IconWindowHint | IconPositionHint | WindowGroupHint;
    wmhints.window_group = window_main_;
    wmhints.icon_window = window_icon_;
      
    wmhints.icon_x = size_hints.x;
    wmhints.icon_y = size_hints.y;
  } 
  else 
  {
    wmhints.initial_state = NormalState;
    wmhints.flags = WindowGroupHint | StateHint;
  }

  XSetWMHints(display_, window_main_, &wmhints);

  class_hints.res_name  = NAME;
  class_hints.res_class = CLASS;
  
  XSetClassHint(display_, window_main_, &class_hints);
  XSetClassHint(display_, window_icon_, &class_hints);
   
  
  if (XStringListToTextProperty(&wname, 1, &name) == 0)
  {
    std::cerr << wname << ": can't allocate window name" << std::endl;
    exit(1);
  }
  
  XSetWMName(display_, window_main_, &name);
  XSetCommand(display_, window_main_, argv, argc);
  XSetWMProtocols(display_, window_main_, &deleteWin, 1); // Close
  XStoreName(  display_, window_main_, NAME);
  XSetIconName(display_, window_main_, NAME); 

  XSelectInput(display_, window_main_, ButtonPressMask | ExposureMask | 
	       ButtonReleaseMask | PointerMotionMask | StructureNotifyMask);
  XSelectInput(display_, window_icon_, ButtonPressMask | ExposureMask | 
	       ButtonReleaseMask | PointerMotionMask | StructureNotifyMask);

}

//--------------------------------------------------------------------
void XHandler::initGraphicsContext()
{
  XGCValues gcv;

  gcv.graphics_exposures = false;
  gcv.foreground = fore_pix_;
  gcv.background = back_pix_;

  graphics_context_ = XCreateGC(display_, window_main_, 
				GCForeground | GCBackground | 
				GCGraphicsExposures, &gcv);  
}

//--------------------------------------------------------------------
void XHandler::drawRectangle(unsigned long color, unsigned x, unsigned y,
			     unsigned width, unsigned height)

{
  XSetForeground(display_, graphics_context_, color);
  XFillRectangle(display_, pixmap_display_, graphics_context_, 
		 x, y, width, height);
}

//--------------------------------------------------------------------
void XHandler::drawShade(Pixmap* pixmap, unsigned x, unsigned y,
			 unsigned width, unsigned height)
{
  XCopyArea(display_, *pixmap, pixmap_display_, graphics_context_,
	    0, 0, 1, height-1, x+width-1, y+1);
  XCopyArea(display_, *pixmap, pixmap_display_, graphics_context_,
	    width-1, 1, 1, height-1, x, y);
  XCopyArea(display_, *pixmap, pixmap_display_, graphics_context_,
	    0, 0, width-1, 1, x+1, y+height-1);
  XCopyArea(display_, *pixmap, pixmap_display_, graphics_context_,
	    1, height-1, width-1, 1, x, y);

  XCopyArea(display_, *pixmap, pixmap_display_, 
	    graphics_context_, 1, 1, width-2, height-2, x+1, y+1);
}

//--------------------------------------------------------------------
void XHandler::drawPixmap(Pixmap* pixmap, unsigned x, unsigned y,
			  unsigned width, unsigned height)
{
  XCopyArea(display_, *pixmap, pixmap_display_, 
	    graphics_context_, 0, 0, width, height, x , y );
}

//--------------------------------------------------------------------
void XHandler::drawPixmapOffset(Pixmap* pixmap, unsigned startx, unsigned starty,
				unsigned x, unsigned y,
				unsigned width, unsigned height)
{
  XCopyArea(display_, *pixmap, pixmap_display_, 
	    graphics_context_, startx, starty, width, height, x, y);
}

//--------------------------------------------------------------------
unsigned XHandler::getColor(const char *colorname) const
{
  XColor color;
  XWindowAttributes winattr;

  XGetWindowAttributes(display_, window_root_, &winattr);
  color.pixel=0;
  XParseColor(display_, winattr.colormap, colorname, &color);
  color.flags = DoRed | DoGreen | DoBlue;
  XAllocColor(display_, winattr.colormap, &color);

  return color.pixel;
}

//--------------------------------------------------------------------
int XHandler::getWindowSize() const
{
  return window_size_;
}

//--------------------------------------------------------------------
Atom XHandler::getDeleteWin() const
{
  return deleteWin;
}

//--------------------------------------------------------------------
//Pixmap* XHandler::readPixmapFile(char* file_name, unsigned* width = NULL, unsigned* height = NULL)
Pixmap* XHandler::readPixmapFile(char* file_name, unsigned* width,
				 unsigned* height, Pixmap* mask)
{
  Pixmap* p = new Pixmap;
  XpmAttributes attributes;
  int status;
  attributes.valuemask = 0;

  status = XpmReadFileToPixmap(display_, 
			       window_main_,
			       file_name,
			       p, mask, &attributes);

  if (status != XpmSuccess)
  {
    std::cerr << "XpmError: " << XpmGetErrorString(status) << " ("<< file_name << ")" << std::endl;
    exit(1);
  }

  if(width != NULL)
    *width = attributes.width;


  if(height != NULL)
    *height = attributes.height;

  return p;
}

//--------------------------------------------------------------------
void XHandler::redraw()
{
  //    XSetClipMask(display_, graphics_context_, *pixmap_mask_);
  //    XCopyArea(display_, pixmap_display_, window_icon_,
  //  	    graphics_context_, 0, 0, 64, 64, 0, 0);
  //    XSetClipMask(display_, graphics_context_, None);
  //    XCopyArea(display_, pixmap_display_, window_main_,
  //  	    graphics_context_, 0, 0, 64, 64, 0, 0);  

//   XSetClipMask(display_, graphics_context_, *pixmap_mask_);
//   XCopyArea(display_, pixmap_display_, window_icon_,
// 	    graphics_context_, 0, 0, width_, height_, offset_w_, offset_h_);
//   XSetClipMask(display_, graphics_context_, None);
//   XCopyArea(display_, pixmap_display_, window_main_,
// 	    graphics_context_, 0, 0, width_, height_, offset_w_, offset_h_);  
  
//    XSetClipMask(display_, graphics_context_, *pixmap_mask_);
//    XCopyArea(display_, pixmap_display_, window_icon_,
//  	    graphics_context_, 0, 0, width_, height_, 0, 0);
//   XSetClipMask(display_, graphics_context_, None);
   XCopyArea(display_, pixmap_display_, window_main_,
 	    graphics_context_, 0, 0, width_, height_, offset_w_, offset_h_);  

   //   XSetClipMask(display_, graphics_context_, *pixmap_mask_);
   //   XSetClipMask(display_, graphics_context_, None);
   XCopyArea(display_, pixmap_display_, window_icon_,
 	    graphics_context_, 0, 0, width_, height_, offset_w_, offset_h_);  

}

//--------------------------------------------------------------------
unsigned XHandler::mixColor(const char *colorname1, int prop1, 
			    const char *colorname2, int prop2)
{
  XColor color, color1, color2;
  XWindowAttributes winattr;

  XGetWindowAttributes(display_, window_root_, &winattr);

  XParseColor(display_, winattr.colormap, colorname1, &color1);
  XParseColor(display_, winattr.colormap, colorname2, &color2);

  color.pixel = 0;
  color.red   = (color1.red*prop1+color2.red*prop2)/(prop1+prop2);
  color.green = (color1.green*prop1+color2.green*prop2)/(prop1+prop2);
  color.blue  = (color1.blue*prop1+color2.blue*prop2)/(prop1+prop2);
  color.flags = DoRed | DoGreen | DoBlue;

  XAllocColor(display_, winattr.colormap, &color);

  return color.pixel;
}


